import copy
import logging
import math
import os
from typing import Callable, Optional, Tuple, Union

import glasflow
import matplotlib.pyplot as plt
from scipy import stats
import torch

from .base import ProposalWithPool

from ..flows import get_realnvp
from ..tensorlist import TensorList
from ..transforms import logit_with_jacobian, sigmoid_with_jacobian
from ..utils.sample import (
    sample_nball,
    sample_radially_truncated_gaussian,
)

logger = logging.getLogger(__name__)


class FlowProposal(ProposalWithPool):
    """Proposal that draws samples using a normalizing flow"""

    trainable: bool = True
    flow: glasflow.flows.base.Flow = None
    scale: torch.Tensor = None
    shift: torch.Tensor = None
    flow_config: dict = None

    def __init__(
        self,
        *,
        dims: int,
        device: torch.DeviceObjType,
        log_likelihood_fn: Callable,
        poolsize: int,
        batch_size: int = 50_000,
        logit: bool = False,
        constant_volume_fraction: Optional[float] = None,
        truncate_log_q: bool = False,
        sample_nball: bool = False,
        max_samples: int = 1_000_000,
        flow_config: Optional[dict] = None,
        accumulate_weights: bool = False,
    ) -> None:
        super().__init__(
            dims=dims, device=device, log_likelihood_fn=log_likelihood_fn
        )

        self.poolsize = poolsize
        self.batch_size = batch_size
        self.truncate_log_q = truncate_log_q
        self.sample_nball = sample_nball
        self.logit = logit
        self.max_samples = max_samples
        self.count = 0
        self.accumulate_weights = accumulate_weights

        self.configure_flow(flow_config)
        self.configure_constant_volume_mode(constant_volume_fraction)

    def configure_constant_volume_mode(
        self, constant_volume_fraction: float
    ) -> None:
        """Configure constant volume mode.

        Sets the radius and u_max (is :code:`sample_nball=False`)
        """
        self.constant_volume_fraction = constant_volume_fraction
        self.u_max = None
        if self.constant_volume_fraction is not None:
            self.cvm_radius = torch.tensor(
                stats.chi.ppf(self.constant_volume_fraction, self.dims)
            )
            if not self.sample_nball:
                self.u_max = torch.tensor(self.constant_volume_fraction)
            logger.debug(f"CVM radius: {self.cvm_radius.item():.3f}")
        else:
            self.cvm_radius = None

    def configure_flow(self, config: Union[dict, None]) -> None:
        default_config = dict(
            dims=self.dims,
            n_neurons=max(2 * self.dims, 8),
            batch_norm_between_transforms=True,
        )
        if config:
            default_config.update(config)
        self.flow_config = default_config

    def train(
        self,
        samples: torch.Tensor,
        logl: torch.Tensor,
        reset: bool = True,
        validation_fraction: float = 0.2,
        patience: int = 20,
    ) -> None:
        if reset or self.flow is None:
            logger.debug("Resetting flow")
            self.flow = get_realnvp(**self.flow_config)
            self.flow = self.flow.to(self.device)

        logger.debug("Training flowproposal")

        max_epochs = 500
        optimizer = torch.optim.Adam(
            self.flow.parameters(),
            lr=5e-3,
            weight_decay=1e-6,
        )

        if self.logit:
            f = torch.logit
        else:
            f = lambda x: x

        self.scale = f(samples).std(dim=0)
        self.shift = f(samples).mean(dim=0)
        samples_prime, _ = self.rescale(samples)

        indices = torch.randperm(len(samples))
        samples_prime = samples_prime[indices, ...]

        n_train = int((1 - validation_fraction) * len(samples))
        train_samples, val_samples = (
            samples_prime[:n_train, ...],
            samples_prime[n_train:, ...],
        )

        history = dict(
            loss=TensorList(buffer_size=max_epochs, device=self.device),
            val_loss=TensorList(buffer_size=max_epochs, device=self.device),
        )
        best_val_loss = torch.inf
        best_it = 0
        best_state = None

        for it in range(max_epochs):
            optimizer.zero_grad()
            loss = -self.flow.log_prob(train_samples).mean()
            loss.backward()

            torch.nn.utils.clip_grad_norm_(self.flow.parameters(), 10)

            optimizer.step()
            history["loss"].append(loss.detach())

            if not torch.isfinite(loss):
                raise RuntimeError("Non-finite loss value")

            with torch.inference_mode():
                val_loss = -self.flow.log_prob(val_samples).mean()
            history["val_loss"].append(val_loss.detach())

            if val_loss < best_val_loss:
                best_val_loss = val_loss
                best_it = it
                best_state = copy.deepcopy(self.flow.state_dict())

            if (it - best_it) > patience:
                logger.debug("Stopping training early")
                break

        if best_state:
            self.flow.load_state_dict(best_state)
        else:
            raise RuntimeWarning("Flow failed to train!")
        self.flow.eval()

    def rescale(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        log_j = -torch.log(self.scale).sum() * torch.ones(
            x.shape[0], device=x.device
        )
        if self.logit:
            x, lj = logit_with_jacobian(x)
            log_j += lj.sum(-1)
        x = (x - self.shift) / self.scale
        return x, log_j

    def rescale_inverse(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        log_j = torch.log(self.scale).sum() * torch.ones(
            x.shape[0], device=x.device
        )
        x = self.scale * x + self.shift
        if self.logit:
            x, lj = sigmoid_with_jacobian(x)
            log_j += lj.sum(-1)
        return x, log_j

    def _draw_latent_samples(
        self, batch_size: int
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if self.cvm_radius:
            if self.sample_nball:
                z = sample_nball(
                    dims=self.dims,
                    n=batch_size,
                    radius=self.cvm_radius,
                    device=self.device,
                )
                log_q = torch.zeros(batch_size, device=self.device)
            else:
                z = sample_radially_truncated_gaussian(
                    dims=self.dims,
                    n=batch_size,
                    radius=self.cvm_radius,
                    device=self.device,
                    u_max=self.u_max,
                )
                # log_q is not normalized but that does not matter
                # because we normalize the weights in the rejection sampling
                log_q = self.flow._distribution.log_prob(z)
        else:
            z, log_q = self.flow._distribution.sample_and_log_prob(batch_size)
        return z, log_q

    def populate(
        self,
        live_points: torch.Tensor,
        logl: torch.Tensor,
        n: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> None:
        if n is None:
            n = self.poolsize

        if batch_size is None:
            batch_size = self.batch_size

        samples = torch.empty((0, self.dims), device=self.device)
        log_weights = torch.empty(0, device=self.device)
        log_n = math.log(n)
        log_m = -math.inf
        log_constant = -torch.inf
        n_accepted = 0
        n_proposed = 0

        if self.truncate_log_q:
            x, log_j = self.rescale(live_points)
            log_q_min = (self.flow.forward_and_log_prob(x)[1] + log_j).min()
        else:
            log_q_min = None

        self.flow.eval()

        while n_accepted < n:
            n_proposed += batch_size
            with torch.inference_mode():
                z, log_q = self._draw_latent_samples(batch_size)
                x, log_j = self.flow.inverse(z)
            log_q = log_q - log_j

            # Rescale
            x, log_j = self.rescale_inverse(x)
            log_q = log_q - log_j
            # Reject out-of-bounds
            ib = ~torch.any((x < 0.0) | (x > 1.0), dim=1)
            if not torch.any(ib):
                continue
            x, log_q = x[ib, ...], log_q[ib]

            if log_q_min:
                keep = log_q > log_q_min
                x, log_q = x[keep, ...], log_q[keep]

            # Rejection sampling
            # w = p / q
            log_w = -log_q

            if self.accumulate_weights:
                log_weights = torch.cat([log_weights, log_w])
                log_constant = max(log_constant, torch.max(log_w))
                samples = torch.cat([samples, x], dim=0)
                log_m = torch.logsumexp(log_weights - log_constant, -1)
                if log_m >= log_n:
                    log_u = torch.log(torch.rand_like(log_weights))
                    accept = (log_weights - log_constant) > log_u
                    n_accepted = accept.sum()
                if len(samples) >= self.max_samples:
                    logger.warning(
                        "Reached max samples (%s)", self.max_samples
                    )
                    log_u = torch.log(torch.rand_like(log_weights))
                    accept = (log_weights - log_constant) > log_u
                    n_accepted = accept.sum()
                    break
            else:
                log_w = log_w - log_w.max()
                log_u = torch.log(torch.rand_like(log_w))
                accept = log_w > log_u
                n_accept_batch = accept.sum()
                m = min(n - n_accepted, n_accept_batch)
                samples = torch.cat([samples, x[accept][:m]])
                n_accepted += m

        acceptance = n_accepted / n_proposed

        if self.accumulate_weights:
            self.samples = samples[accept][:n]
        else:
            self.samples = samples[:n]
        self.logl = None
        self.indices = list(range(len(self.samples)))
        self.count += 1
        self.populated = True
        return acceptance



from nessai_torch.utils.sample import sample_radially_truncated_gaussian
import pytest
from scipy import stats
import torch


@pytest.fixture(params=[2, 4, 8])
def dims(request):
    return request.param


@pytest.fixture(params=[1, 10, 100])
def n(request):
    return request.param


@pytest.fixture(params=torch.tensor([1.0, 4.0, 16.0, 64.0]))
def radius(request):
    return request.param


def test_truncated_gaussian_radius(dims, n, radius, device):
    x = sample_radially_truncated_gaussian(
        dims=dims, n=n, radius=radius, device=device
    )
    r = torch.sqrt(torch.sum(x**2, dim=-1))
    assert torch.all(r <= radius)


def test_truncated_gaussian_u_max(dims, n, radius, device):
    u_max = stats.chi(df=dims).cdf(radius.item())
    x = sample_radially_truncated_gaussian(
        dims=dims, n=n, u_max=u_max, device=device
    )
    r = torch.sqrt(torch.sum(x**2, dim=-1))
    assert torch.all(r <= radius)



from typing import Optional
import torch

from .utils.maths import log_integrate_log_trap, logsubexp
from .tensorlist import TensorList


class EvidenceIntegral:
    def __init__(self, *, nlive: int, device: torch.DeviceObjType):
        self.device = device
        self.nlive = nlive
        self.logt = torch.tensor(
            -1.0 / nlive, requires_grad=False, device=device
        )
        self.logz = -torch.tensor(
            torch.inf, requires_grad=False, device=device
        )
        self.logw = torch.tensor(0.0, requires_grad=False, device=device)
        self.logx = TensorList(device=device)
        self.logx.append(torch.tensor(0.0, requires_grad=False, device=device))
        self.logl = TensorList(device=device)
        self.logl.append(
            torch.tensor(-torch.inf, requires_grad=False, device=device)
        )
        self.info = torch.tensor(0.0, requires_grad=False, device=device)

    def update(self, logl: torch.Tensor, nlive: Optional[int] = None):
        """Update the integral"""
        logz_current = self.logz.clone()
        if nlive is not None:
            logt = torch.tensor(-1.0 / nlive, device=self.device)
        else:
            logt = self.logt
        logz = self.logw + torch.log1p(-torch.exp(self.logt)) + logl
        self.logz = torch.logaddexp(self.logz, logz)
        self.info = (
            torch.exp(logz - self.logz) * logl
            + torch.exp(logz_current - self.logz) * (self.info + logz_current)
            - self.logz
        )
        if torch.isnan(self.info):
            self.info = torch.tensor(
                0.0, requires_grad=False, device=self.device
            )
        self.logw += logt
        self.logx.append(self.logw.clone())
        self.logl.append(logl)

    @property
    def log_posterior_weights(self) -> torch.Tensor:
        """Compute the log-posterior weights."""
        logl = torch.cat((self.logl.data, self.logl[-1].view(1)), dim=0)
        logx = torch.cat(
            (
                self.logx.data,
                torch.tensor(-torch.inf, device=self.device).view(1),
            ),
            dim=0,
        )
        logz = log_integrate_log_trap(logl, logx)
        logw = logsubexp(logx[:-1], logx[1:])
        log_post_w = logl[1:-1] + logw[:-1] - logz
        return log_post_w

    @property
    def logz_error(self) -> torch.Tensor:
        return torch.sqrt(self.info / self.nlive)



import glasflow
from glasflow import RealNVP
import torch


def get_realnvp(
    *,
    dims: int,
    n_neurons: int,
    n_transforms: int = 4,
    n_layers: int = 1,
    batch_norm_between_transforms: bool = True,
    compile: bool = False,
    **kwargs,
) -> glasflow.flows.base.Flow:
    flow = RealNVP(
        n_inputs=dims,
        n_transforms=n_transforms,
        n_blocks_per_transform=n_layers,
        n_neurons=n_neurons,
        batch_norm_between_transforms=batch_norm_between_transforms,
        **kwargs,
    )
    if compile:
        flow = torch.compile(flow)
    return flow



import logging


def configure_logger(log_level="INFO"):
    """Configure the logger"""
    logger = logging.getLogger("nessai_torch")
    logger.setLevel(log_level)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)
    logger.addHandler(stream_handler)
    return logger



from abc import abstractmethod
from typing import Callable, List, Optional

import torch


class Proposal:
    """ "Base class for all proposals."""

    has_pool: bool = False
    """Indicates if the proposal has a pool of samples."""
    trainable: bool = False
    """Indicates if the proposal can be trained."""

    def __init__(
        self,
        *,
        dims: int,
        device: torch.DeviceObjType,
        log_likelihood_fn: Callable,
    ) -> None:
        """
        Parameters
        ----------
        dims
            Dimensionality of the parameter space.
        device
            Torch device.
        log_likelihood_fn
            Function for evaluating the log-likelihood for samples in the
            unit hypercube space.
        """
        self.dims = dims
        self.device = device
        self.log_likelihood_fn = log_likelihood_fn

    @abstractmethod
    def draw(self, x: torch.Tensor) -> None:
        raise NotImplementedError


class ProposalWithPool(Proposal):
    has_pool: bool = True
    populated: bool = False
    trainable: bool = False
    indices: List[int] = None
    samples: torch.Tensor = None
    logl: torch.Tensor = None

    @abstractmethod
    def populate(self, live_points: torch.Tensor, logl: torch.Tensor) -> None:
        raise NotImplementedError

    def compute_likelihoods(
        self, log_likelihood_fn: Optional[Callable] = None
    ) -> None:
        """Compute the log-likelihoods for the pool of samples.

        Parameters
        ----------
        log_likelihood_fn
            Function to use instead of the function that was specified when
            instantiating the class.
        """
        if self.samples is None or not self.populated:
            raise RuntimeError("Proposal is not populated")
        if log_likelihood_fn is None:
            log_likelihood_fn = self.log_likelihood_fn
        self.logl = log_likelihood_fn(self.samples)

    def draw(self, x: torch.Tensor) -> torch.Tensor:
        if not self.populated:
            raise RuntimeError(
                "Proposal must be populated before drawing a sample"
            )
        index = self.indices.pop()
        new_sample = self.samples[index]
        if not self.indices:
            self.populated = False
            self.samples = None
        if self.logl is None:
            new_logl = None
        else:
            new_logl = self.logl[index]
        return new_sample, new_logl



"""Implementation of the inverse incomplete gamma function.

Implementation is based directly of the implementation in tensorflow
probability: https://www.tensorflow.org/probability/api_docs/python/tfp/math/igammainv
"""
import math
import torch
from torch.distributions.utils import euler_constant as euler_gamma
from typing import List


@torch.jit.script
def _didonato_eq_twenty_three(
    log_b: torch.Tensor, v: torch.Tensor, a: torch.Tensor
) -> torch.Tensor:
    return (
        -log_b + torch.xlogy(a - 1.0, v) - torch.log1p((1.0 - a) / (1.0 + v))
    )


@torch.jit.script
def _didonato_eq_twenty_five(a: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Compute Equation 25 from Didonato's paper."""
    c1 = torch.xlogy(a - 1.0, y)
    c1_sq = torch.square(c1)
    c1_cub = c1_sq * c1
    c1_fourth = torch.square(c1_sq)
    a_sq = torch.square(a)
    a_cub = a_sq * a
    c2 = (a - 1.0) * (1.0 + c1)
    c3 = (a - 1.0) * ((3.0 * a - 5.0) / 2.0 + c1 * (a - 2.0 - c1 / 2.0))
    c4 = (a - 1.0) * (
        (c1_cub / 3.0)
        - (3.0 * a - 5.0) * c1_sq / 2.0
        + (a_sq - 6.0 * a + 7.0) * c1
        + (11.0 * a_sq - 46.0 * a + 47.0) / 6.0
    )
    c5 = (a - 1.0) * (
        -c1_fourth / 4.0
        + (11.0 * a - 17.0) * c1_cub / 6
        + (-3.0 * a_sq + 13.0 * a - 13.0) * c1_sq
        + (2.0 * a_cub - 25.0 * a_sq + 72.0 * a - 61.0) * c1 / 2.0
        + (25.0 * a_cub - 195.0 * a_sq + 477 * a - 379) / 12.0
    )
    return y + c1 + (((c5 / y + c4) / y + c3 / y) + c2) / y


@torch.jit.script
def _polyval(coeffs: List[float], x: torch.Tensor) -> torch.Tensor:
    val = torch.zeros_like(x)
    for c in coeffs[:-1]:
        val = (val + c) * x
    return val + coeffs[-1]


@torch.jit.script
def _didonato_eq_thirty_two(p: torch.Tensor, q: torch.Tensor) -> torch.Tensor:
    """Compute Equation 32 from Didonato's paper."""
    numerator_coeffs = [
        0.213623493715853,
        4.28342155967104,
        11.6616720288968,
        3.31125922108741,
    ]
    denominator_coeffs = [
        0.36117081018842e-1,
        1.27364489782223,
        6.40691597760039,
        6.61053765625462,
        1.0,
    ]
    t = torch.where(
        p < 0.5, torch.sqrt(-2 * torch.log(p)), torch.sqrt(-2.0 * torch.log(q))
    )
    result = t - _polyval(numerator_coeffs, t) / _polyval(
        denominator_coeffs, t
    )
    return torch.where(p < 0.5, -result, result)


def _didonato_eq_thirty_four(a, x):
    """Compute Equation 34 from Didonato's paper."""
    # This function computes `S_n` in equation thirty four.

    tolerance = 1e-4

    def _taylor_series(should_stop, index, partial, series_sum):
        partial = partial * x / (a + index)
        series_sum = torch.where(should_stop, series_sum, series_sum + partial)
        should_stop = (partial < tolerance) | (index > 100)
        return should_stop, index + 1, partial, series_sum

    partial = torch.ones_like(a + x)
    series_sum = torch.ones_like(a + x)
    should_stop = torch.zeros_like(a + x, dtype=torch.bool)
    index = 1

    # Max 100 iterations
    while torch.any(~should_stop):
        should_stop, index, partial, series_sum = _taylor_series(
            should_stop, index, partial, series_sum
        )

    return series_sum


def _inverse_gammainc_initial_approx(
    a: torch.Tensor,
    p: torch.Tensor,
    q: torch.Tensor,
    use_p_for_logq: bool = True,
) -> torch.Tensor:
    """Compute an initial guess for :code:`gammaincinv(a, p)`."""
    lgamma_a = torch.lgamma(a)

    # This ensures that computing log(1 - p) avoids roundoff errors. This is
    # needed since igammacinv and igammainv both use this codepath,
    if use_p_for_logq:
        log_q = torch.log1p(-p)
    else:
        log_q = torch.log(q)

    log_b = log_q + lgamma_a

    result = _didonato_eq_twenty_five(a, -log_b)

    # The code below is for when a < 1.

    v = -log_b - (1.0 - a) * torch.log(-log_b)
    v_sq = torch.square(v)

    # This is Equation 24.
    result = torch.where(
        log_b > math.log(0.01),
        -log_b
        - (1.0 - a) * torch.log(v)
        - torch.log(
            (v_sq + 2.0 * (3.0 - a) * v + (2.0 - a) * (3 - a))
            / (v_sq + (5.0 - a) * v + 2.0)
        ),
        result,
    )

    result = torch.where(
        log_b >= math.log(0.15), _didonato_eq_twenty_three(log_b, v, a), result
    )

    t = torch.exp(-euler_gamma - torch.exp(log_b))
    u = t * torch.exp(t)
    result = torch.where(
        (a < 0.3) & (log_b >= math.log(0.35)), t * torch.exp(u), result
    )

    # These are hand tuned constants to compute (p * Gamma(a + 1)) ** (1 / a)
    # This is Equation 21.
    u = torch.where(
        (torch.exp(log_b) * q > 1e-8) & (q > 1e-5),
        torch.pow(p * torch.exp(lgamma_a) * a, torch.reciprocal(a)),
        # When (1 - p) * Gamma(a) or (1 - p) is small,
        # we can taylor expand Gamma(a + 1) ** 1 / a to get
        # exp(-euler_gamma for the zeroth order term.
        # Also p ** 1 / a = exp(log(p) / a) = exp(log(1 - q) / a)
        # ~= exp(-q / a) resulting in the following expression.
        torch.exp((-q / a) - euler_gamma),
    )

    result = torch.where(
        (log_b > math.log(0.6)) | ((log_b >= math.log(0.45)) & (a >= 0.3)),
        u / (1.0 - (u / (a + 1.0))),
        result,
    )

    # The code below is for when a < 1.

    sqrt_a = torch.sqrt(a)
    s = _didonato_eq_thirty_two(p, q)
    s_sq = torch.square(s)
    s_cub = s_sq * s
    s_fourth = torch.square(s_sq)
    s_fifth = s_fourth * s

    # This is the Cornish-Fisher 6 term expansion for x (by viewing igammainv as
    # the quantile function for the Gamma distribution). This is equation (31).
    w = a + s * sqrt_a + (s_sq - 1.0) / 3.0
    w = w + (s_cub - 7.0 * s) / (36.0 * sqrt_a)
    w = w - (3.0 * s_fourth + 7.0 * s_sq - 16.0) / (810 * a)
    w = w + (9.0 * s_fifth + 256.0 * s_cub - 433.0 * s) / (38880 * a * sqrt_a)

    # The code below is for when a > 1. and p > 0.5.
    d = torch.maximum(torch.tensor(2.0), a * (a - 1.0))
    result_a_large_p_large = torch.where(
        log_b <= -d * math.log(10.0),
        _didonato_eq_twenty_five(a, -log_b),
        _didonato_eq_twenty_three(
            log_b, _didonato_eq_twenty_three(log_b, w, a), a
        ),
    )
    result_a_large_p_large = torch.where(
        w < 3.0 * a, w, result_a_large_p_large
    )
    result_a_large_p_large = torch.where(
        (a >= 500.0) & (torch.abs(1.0 - w / a) < 1e-6),
        w,
        result_a_large_p_large,
    )

    # The code below is for when a > 1. and p <= 0.5.
    z = w
    v = torch.log(p) + torch.lgamma(a + 1.0)

    # The code below follows Equation 35 which involves multiple evaluations of
    # F_i.
    modified_z = torch.exp((v + w) / a)
    for _ in range(2):
        s = torch.log1p(
            modified_z / (a + 1.0) * (1.0 + modified_z / (a + 2.0))
        )
        modified_z = torch.exp((v + modified_z - s) / a)

    s = torch.log1p(
        modified_z
        / (a + 1.0)
        * (1.0 + modified_z / (a + 2.0) * (1.0 + modified_z / (a + 3.0)))
    )
    modified_z = torch.exp((v + modified_z - s) / a)
    z = torch.where(w <= 0.15 * (a + 1.0), modified_z, z)

    ls = torch.log(_didonato_eq_thirty_four(a, z))
    medium_z = torch.exp((v + z - ls) / a)
    result_a_large_p_small = torch.where(
        (z <= 0.01 * (a + 1.0)) | (z > 0.7 * (a + 1.0)),
        z,
        medium_z
        * (
            1.0
            - (a * torch.log(medium_z) - medium_z - v + ls) / (a - medium_z)
        ),
    )

    result_a_large = torch.where(
        p <= 0.5, result_a_large_p_small, result_a_large_p_large
    )
    result = torch.where(a < 1.0, result, result_a_large)

    # This ensures that computing log(1 - p) avoids roundoff errors. This is
    # needed since igammacinv and igammainv both use this codepath,
    # switching p and q.
    result = torch.where(torch.eq(a, torch.tensor(1.0)), -log_q, result)
    return result


def _shared_igammainv_computation(
    a: torch.Tensor, p: torch.TensorType, is_gammaincinv: bool = True
):
    """Computation for gammaincinv."""

    if is_gammaincinv:
        q = 1.0 - p
    else:
        q = p
        p = 1.0 - q

    x = _inverse_gammainc_initial_approx(
        a, p, q, use_p_for_logq=is_gammaincinv
    )

    # Run 3 steps of Newton-Halley method.
    for _ in range(3):
        factorial = torch.exp(a * torch.log(x) - x - torch.lgamma(a))

        f_over_der = torch.where(
            ((p <= 0.9) & is_gammaincinv) | ((q > 0.9) & (not is_gammaincinv)),
            (torch.igamma(a, x) - p) * x / factorial,
            -(torch.igammac(a, x) - q) * x / factorial,
        )
        second_der_over_der = -1.0 + (a - 1.0) / x
        modified_x = torch.where(
            ~torch.isfinite(second_der_over_der),
            # Use Newton's method if the second derivative is not available.
            x - f_over_der,
            # Use Halley's method otherwise. Halley's method is:
            # x_{n+1} = x_n - f(x_n) / f'(x_n) * (
            #    1 - f(x_n) / f'(x_n) * 0.5 f''(x_n) / f'(x_n))
            x - f_over_der / (1.0 - 0.5 * f_over_der * second_der_over_der),
        )
        x = torch.where(torch.eq(factorial, 0.0), x, modified_x)
    x = torch.where((a < 0.0) | (p < 0.0) | (p > 1.0), torch.nan, x)
    x = torch.where(torch.eq(p, 0.0), 0.0, x)
    x = torch.where(torch.eq(p, 1.0), torch.inf, x)

    return x


def gammaincinv(a: torch.Tensor, p: torch.Tensor) -> torch.Tensor:
    """Inverse of the incomplete gamma function.

    Derivatives are not implemented. This function cannot be JIT compiled.
    """
    return _shared_igammainv_computation(a, p, is_gammaincinv=True)



from typing import Optional

from scipy import stats
import torch
from .gammaincinv import gammaincinv


def rejection_sample(
    samples: torch.Tensor, log_w: torch.Tensor
) -> torch.Tensor:
    """Perform rejection sampling given samples and log-weights."""
    log_u = torch.log(torch.rand_like(log_w))
    log_w = log_w - log_w.max()
    return samples[log_w > log_u]


def sample_nball(
    dims: int,
    n: int = 1,
    radius: float = 1.0,
    device: Optional[torch.DeviceObjType] = None,
) -> torch.Tensor:
    """Sample from an n-ball with radius r"""
    z = torch.randn(n, dims, device=device)
    z = z / torch.sqrt(torch.sum(z**2, dim=1, keepdim=True))
    r = radius * (torch.rand(n, 1, device=device) ** (1 / dims))
    return r * z


def sample_radially_truncated_gaussian(
    dims: int,
    n: int = 1,
    radius: float = 1.0,
    u_max: Optional[float] = None,
    device: Optional[torch.DeviceObjType] = None,
) -> torch.Tensor:
    """Sample from a radially truncated Gaussian.

    Truncation can be specified with a radius or the maximum value of the
    CDF (:code:`u_max`)

    Parameters
    ----------
    dims
        The number of dimensions.
    n
        The number of samples to draw.
    radius
        The radius at which to truncate the Gaussian. Ignored if :code:`u_max`
        is specified.
    device
        The device to use for the samples.

    Returns
    -------
    Tensor of samples with shape (n, dims)
    """
    if u_max is None:
        u_max = stats.chi(df=dims).cdf(radius.cpu().numpy())
    u = u_max * torch.rand(n, device=device)
    # Inverse CDF of a chi-distribution
    p = torch.sqrt(
        2 * gammaincinv(torch.tensor(0.5 * dims, device=u.device), u)
    )
    x = torch.randn(dims, n, device=device)
    points = (p * x / torch.sqrt(torch.sum(torch.square(x), dim=0))).T
    return points



"""Basic example using the Rosenbrock likelihood."""

import os

import torch

from nessai_torch.sampler import Sampler
from nessai_torch.plot import corner_plot
from nessai_torch.utils.bounds import in_bounds
from nessai_torch.utils.logging import configure_logger

# Limit torch to a single thread
torch.set_num_threads(1)
torch.set_num_interop_threads(1)

logger = configure_logger("INFO")

dims = 4
bounds = torch.tensor([-5.0, 5.0])
outdir = "outdir_rosenbrock"


@torch.jit.script
def rosenbrock(x: torch.Tensor) -> torch.Tensor:
    r"""Rosenbrock function in N dimensions.

    This is the more involved variant given by

    .. math::
        \sum_{i=1}^{N-1} [100(x_{i+1} - x_{i}^{2})^2 + (1 - x_{i})^2].
    """
    return torch.sum(
        100.0 * (x[..., 1:] - x[..., :-1] ** 2.0) ** 2.0
        + (1.0 - x[..., :-1]) ** 2.0,
        dim=-1,
    )


# Define the log-likelihood
def log_likelihood(x: torch.Tensor) -> torch.Tensor:
    return torch.log(in_bounds(x, bounds)) - rosenbrock(x)


# Define the mapping from the unit hyper-cube to the prior
def prior_transform(x: torch.Tensor) -> torch.Tensor:
    return (bounds[1] - bounds[0]) * x + bounds[0]


# Initialise the sampler
sampler = Sampler(
    log_likelihood=log_likelihood,
    prior_transform=prior_transform,
    dims=dims,
    outdir=outdir,
    nlive=1_000,
    constant_volume_fraction=0.98,
    flow_config=dict(n_neurons=32, batch_norm_between_transforms=True),
    reset_flow=4,
)

# Run the sampler
sampler.run()

# Plot the posterior samples
corner_plot(
    sampler.posterior_samples, filename=os.path.join(outdir, "posterior.png")
)



"""Proposal that samples from the prior subject to any constraints"""
import torch
from typing import Callable, Tuple

from .base import Proposal


class PriorProposal(Proposal):
    """Proposal that samples from the prior subject to any constraints.

    Evaluates the likelihood of samples
    """

    def draw(self, x) -> Tuple[torch.Tensor, torch.Tensor]:
        """Draw a valid sample from the prior"""
        logl = torch.tensor(-torch.inf)
        while not torch.isfinite(logl):
            x = torch.rand(1, self.dims, device=self.device)
            logl = self.log_likelihood_fn(x)
        return x, logl



"""nessai-torch

Implementation of nessai in PyTorch.
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass


from .sampler import Sampler

__all__ = [
    "Sampler",
]



"""Utilities for handling bounds"""
import torch


@torch.jit.script
def in_bounds(x: torch.Tensor, bounds: torch.Tensor) -> torch.Tensor:
    return ~(
        torch.any(x < bounds[0], dim=-1) | torch.any(x > bounds[1], dim=-1)
    )



from typing import Optional, Tuple

import torch


@torch.jit.script
def logit_with_jacobian(
    x: torch.Tensor, eps: Optional[float] = None
) -> Tuple[torch.Tensor, torch.Tensor]:
    if eps is not None:
        x = torch.clamp(x, eps, 1 - eps)
    log_j = -torch.log(x) - torch.log1p(-x)
    return torch.logit(x), log_j


@torch.jit.script
def sigmoid_with_jacobian(
    x: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    x = torch.sigmoid(x)
    log_j = torch.log(x) + torch.log1p(-x)
    return x, log_j



"""Plotting functions"""
import logging
from typing import List, Optional

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats
import torch

logger = logging.getLogger(__name__)


def save_figure(
    figure: Figure, filename: Optional[str] = None
) -> Optional[Figure]:
    if filename is not None:
        figure.tight_layout()
        figure.savefig(filename)
        plt.close(figure)
    else:
        return figure


def get_default_figsize() -> list:
    return list(plt.rcParams["figure.figsize"])


def plot_trace(
    logx: torch.Tensor,
    nested_samples: torch.Tensor,
    labels: Optional[List[str]] = None,
    filename: Optional[str] = None,
) -> Optional[Figure]:
    """Produce a trace plot."""

    logx = logx.cpu().numpy()
    nested_samples = nested_samples.cpu().numpy()

    dims = nested_samples.shape[-1]

    figsize = (5, dims * 2)

    fig, axs = plt.subplots(dims, 1, sharex=True, figsize=figsize)

    if labels is None:
        labels = [f"x_{i}" for i in range(dims)]

    for i, samples in enumerate(nested_samples.T):
        axs[i].plot(logx, samples, ls="", marker=",")
        axs[i].set_ylabel(labels[i])

    axs[-1].set_xlabel(r"$\log X$")
    axs[-1].invert_xaxis()

    return save_figure(fig, filename)


def plot_insertion_indices(
    indices: torch.Tensor,
    nlive: int,
    confidence_intervals: Optional[List[float]] = None,
    filename: Optional[str] = None,
) -> Optional[Figure]:
    """Plot the insertion indices"""
    indices = indices.cpu().numpy()

    x = np.arange(0, nlive, 1)
    analytic = x / x[-1]

    n = len(indices)
    counts = np.bincount(indices, minlength=nlive)
    estimated = np.cumsum(counts) / n

    if confidence_intervals is None:
        confidence_intervals = [
            0.997,
        ]

    fig, axs = plt.subplots(2, 1, sharex=True)

    axs[0].hist(indices, bins="auto", histtype="step")

    axs[1].plot(x, analytic - estimated)

    for ci in confidence_intervals:
        bound = (1 - ci) / 2
        bound_values = stats.binom.ppf(1 - bound, n, analytic) / n
        lower = bound_values - analytic
        upper = analytic - bound_values
        upper[0] = 0
        upper[-1] = 0
        lower[0] = 0
        lower[-1] = 0
        axs[1].fill_between(x, lower, upper, color="grey", alpha=0.2)

    axs[1].set_xlabel("Insertion index")
    axs[1].set_ylabel("Analytic - estimated")

    axs[1].set_xlim(0, nlive - 1)

    return save_figure(fig, filename)


def corner_plot(
    samples: torch.Tensor,
    labels: Optional[list[str]] = None,
    filename: Optional[str] = None,
    **kwargs,
) -> Optional[Figure]:
    try:
        import corner
    except ImportError as e:
        logger.error(
            f"Could not import corner, skipping corner plot! Error: {e}"
        )
        return
    samples = samples.cpu().numpy()
    fig = corner.corner(samples, labels=labels, **kwargs)
    return save_figure(fig, filename)


def plot_samples_1d(
    *samples: torch.Tensor,
    labels: Optional[list[str]] = None,
    parameter_labels: Optional[list[str]],
    filename: Optional[str] = None,
    **kwargs,
) -> Optional[Figure]:
    """Plot 1d histograms for one or more sets of samples."""
    hist_kwargs = dict(
        density=True,
        histtype="step",
    )
    if kwargs:
        hist_kwargs.update(kwargs)

    if not labels:
        labels = [f"samples_{i}" for i in range(len(samples))]

    dims = samples[0].shape[-1]
    figsize = get_default_figsize()
    figsize[0] /= 2
    figsize[1] *= dims / 2

    fig, axs = plt.subplots(dims, 1, figsize=figsize)

    for array, label in zip(samples, labels):
        for ax, array_1d in zip(axs, array.T):
            ax.hist(array_1d.detach(), label=label, **hist_kwargs)
    axs[-1].legend()

    if parameter_labels:
        for ax, label in zip(axs, parameter_labels):
            ax.set_xlabel(label)

    return save_figure(fig, filename)



import numpy as np


def rolling_mean_numpy(x: np.ndarray, n: int = 10) -> np.ndarray:
    """Compute the rolling mean with a given window size.

    Based on the version in nessai.

    Parameters
    ----------
    x :
        Array of samples
    n :
        Size of the window over which the rolling mean is computed.

    Returns
    -------
    Array containing the rolling mean.
    """
    # np.cumsum is faster but doesn't work with infs in the data.
    return np.convolve(
        np.pad(x, (n // 2, n - 1 - n // 2), mode="edge"),
        np.ones(n) / n,
        mode="valid",
    )



from nessai_torch.utils.gammaincinv import gammaincinv
import pytest
from scipy.special import gammaincinv as gammaincinv_scipy
import torch

# Call once to compile
gammaincinv(torch.tensor(1.0), torch.tensor(1.0))


# The tests fail outside these limits
@pytest.mark.parametrize("a", torch.linspace(1.5, 35, 32))
def test_gammaincinv_invertible(a):
    x = torch.linspace(0.0, 10.0, 10)
    y = torch.special.gammainc(a, x)
    x_out = gammaincinv(a, y)
    assert torch.isclose(x_out, x).all()


@pytest.mark.parametrize("a", torch.linspace(0.1, 101, 50))
def test_gammaincinv_scipy(a):
    y = torch.linspace(0, 1, 1000)
    x_out = gammaincinv(a, y)
    x_expected = gammaincinv_scipy(a.numpy(), y.numpy())
    assert torch.isclose(x_out, torch.tensor(x_expected)).all()



import pytest
import torch

from nessai_torch.sampler import Sampler
from nessai_torch.utils.bounds import in_bounds
from nessai_torch.proposal.flow import FlowProposal
from nessai_torch.proposal.prior import PriorProposal
from nessai_torch.proposal.spherical import SphericalProposal


@pytest.fixture
def dims():
    return 2


@pytest.fixture
def sampler_inputs(dims, device):
    mean = torch.zeros(dims, device=device)
    cov = torch.eye(dims, device=device)
    bounds = torch.tensor([-10.0, 10.0], device=device)
    dist = torch.distributions.MultivariateNormal(mean, cov)

    def log_likelihood(x: torch.tensor) -> torch.tensor:
        return dist.log_prob(x) + torch.log(in_bounds(x, bounds))

    def prior_transform(x: torch.tensor) -> torch.tensor:
        return (bounds[1] - bounds[0]) * x + bounds[0]

    return dims, log_likelihood, prior_transform


@pytest.mark.parametrize(
    "ProposalClass, kwargs",
    [
        (
            FlowProposal,
            {"sample_nball": False, "constant_volume_fraction": 0.95},
        ),
        (
            FlowProposal,
            {"sample_nball": True, "constant_volume_fraction": 0.95},
        ),
        (FlowProposal, {}),
        (SphericalProposal, {}),
        (PriorProposal, {}),
    ],
)
def test_sampling(sampler_inputs, device, ProposalClass, tmp_path, kwargs):
    dims, log_likelihood, prior_transform = sampler_inputs

    outdir = tmp_path / "test_proposal"
    outdir.mkdir()

    sampler = Sampler(
        log_likelihood=log_likelihood,
        prior_transform=prior_transform,
        dims=dims,
        device=device,
        outdir=outdir,
        nlive=50,
        tolerance=5.0,
        proposal_class=ProposalClass,
        **kwargs,
    )
    sampler.run()



import math

import torch


@torch.jit.script
def logsubexp(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    m = torch.max(a, b)
    return m + ((a - m).exp() - (b - m).exp()).log()


@torch.jit.script
def log_integrate_log_trap(
    logf: torch.Tensor, logx: torch.Tensor
) -> torch.Tensor:
    """
    Trapezoidal integration of given log(f). Returns log of the integral.

    Parameters
    ----------
    log_func : array_like
        Log values of the function to integrate over.
    log_support : array_like
        Log prior-volumes for each value.

    Returns
    -------
    float
        Log of the result of the integral.
    """
    logf_sum = torch.logaddexp(logf[:-1], logf[1:]) - math.log(2)
    log_dxs = logsubexp(logx[:-1], logx[1:])
    return torch.logsumexp(logf_sum + log_dxs, 0)



from .base import Proposal

import torch
from typing import Callable


class SphericalProposal(Proposal):
    """Proposal that draws samples from within a sphere with radius given by
    the worst point.

    Parameters
    ----------
    dims
        The number of dimensions.
    device
        The device to use for new samples.
    """

    def __init__(
        self,
        *,
        dims: int,
        device: torch.DeviceObjType,
        log_likelihood_fn: Callable,
        trainable: bool = True,
    ) -> None:
        super().__init__(
            dims=dims, device=device, log_likelihood_fn=log_likelihood_fn
        )
        self.origin = 0.5 * torch.ones(self.dims, device=device)
        self.trainable = trainable

    def train(self, live_points: torch.Tensor, logl: torch.Tensor):
        """Updates the origin to the sample with highest log-likelihood"""
        self.origin = live_points.mean(dim=0)

    def draw(self, sample: torch.tensor, n=1) -> torch.tensor:
        max_radius = torch.sqrt((torch.sum((sample - self.origin) ** 2)))
        z = torch.randn(n, self.dims, device=self.device)
        z = z / torch.sqrt(torch.sum(z**2, dim=1, keepdim=True))
        r = max_radius * (
            torch.rand(n, 1, device=self.device) ** (1 / self.dims)
        )
        x = r * z + self.origin
        return torch.squeeze(x), None



from collections.abc import Sequence
import logging
import math
from typing import Optional, Tuple

import torch

logger = logging.getLogger(__name__)


class TensorList(Sequence):
    def __init__(
        self,
        *,
        device: torch.DeviceObjType = "cpu",
        size: Tuple[int] = None,
        buffer_size: int = 1000,
        dtype: Optional[torch.dtype] = None,
    ):
        self.device = device
        self.buffer_size = buffer_size
        if size is None:
            size = ()
        self._size = size
        self.dtype = dtype if dtype is not None else torch.get_default_dtype()
        self.buffer = self._new_buffer()
        self.current_buffer_size = len(self.buffer)
        self._count = 0

    @property
    def data(self) -> torch.Tensor:
        return self.buffer[: self._count]

    _torch_fn_data: torch.Tensor = data
    """Attribute used for torch functions."""

    def append(self, x: torch.Tensor) -> None:
        self.buffer[self._count] = x
        self._count += 1
        if self._count >= self.current_buffer_size:
            self._extend_buffer()

    @property
    def _remaining(self):
        return len(self.buffer) - self._count

    def extend(self, x: torch.Tensor) -> None:
        n = len(x)
        if n > self._remaining:
            extra = n - self._remaining
            n_buffers = math.ceil(extra / self.buffer_size)
            self._extend_buffer(n_buffers)
        self.buffer[self._count : (self._count + n)] = x
        self._count += n
        if self._count >= self.current_buffer_size:
            self._extend_buffer()

    def _new_buffer(self) -> torch.Tensor():
        return torch.empty(
            self.buffer_size,
            *self._size,
            device=self.device,
            dtype=self.dtype,
        )

    def _extend_buffer(self, n: int = 1) -> None:
        logger.debug("Extending buffer")
        self.buffer = torch.cat(
            [self.buffer] + [self._new_buffer() for _ in range(n)], dim=0
        )
        self.current_buffer_size = len(self.buffer)

    def __getitem__(self, index: int) -> torch.Tensor:
        return self.buffer[index]

    def __len__(self) -> int:
        return self._count

    def __str__(self) -> str:
        return self.data.__str__()

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        args = [
            a._torch_fn_data if hasattr(a, "_torch_fn_data") else a
            for a in args
        ]
        ret = func(*args, **kwargs)
        return ret



import h5py
import numpy as np
import pytest
from unittest.mock import mock_open, patch

from nessai_torch.utils.io import (
    add_dict_to_hdf5_file,
    encode_for_hdf5,
    save_dict_to_hdf5,
)


@pytest.fixture
def data_dict():
    data = dict(
        a=np.array([1, 2, 3]),
        b=np.array([(1, 2)], dtype=[("x", "f4"), ("y", "f4")]),
        cls=object(),
        l=[1, 2, 3],
        dict1={"a": None, "b": 2},
        dict2={"c": [1, 2, 3], "array": np.array([3, 4, 5])},
        s="A string",
        nan=None,
    )
    return data


@pytest.mark.parametrize(
    "value, expected", [(None, "__none__"), ([1, 2], [1, 2])]
)
def test_encode_to_hdf5(value, expected):
    assert encode_for_hdf5(value) == expected


def test_add_dict_to_hdf5_file(tmp_path, data_dict):
    data_dict.pop("cls")
    with h5py.File(tmp_path / "test.h5", "w") as f:
        add_dict_to_hdf5_file(f, "/", data_dict)
        assert list(f.keys()) == sorted(data_dict.keys())
        assert f["/dict1/a"][()].decode() == "__none__"
        np.testing.assert_array_equal(
            f["dict2/array"][:], data_dict["dict2"]["array"]
        )


def test_save_dict_to_hdf5(data_dict):
    f = mock_open()
    filename = "result.h5"
    with patch("h5py.File", f) as mock_file, patch(
        "nessai_torch.utils.io.add_dict_to_hdf5_file"
    ) as mock_add:
        save_dict_to_hdf5(data_dict, filename)
    mock_file.assert_called_once_with(filename, "w")
    mock_add.assert_called_once_with(f(), "/", data_dict)



import matplotlib.pyplot as plt
from nessai_torch.plot import (
    get_default_figsize,
    plot_samples_1d,
)
import pytest
import torch


@pytest.fixture(autouse=True)
def close_plots():
    plt.close("all")


def test_get_default_figsize():
    figsize = get_default_figsize()
    assert len(figsize) == 2


def test_plot_samples_1d():
    samples = [
        torch.randn(10, 2),
        torch.randn(10, 2),
    ]
    plot_samples_1d(
        *samples,
        labels=["train", "val"],
        parameter_labels=["x_0", "x_1"],
    )



"""I/O related utils"""
import h5py
from typing import Any


def encode_for_hdf5(value: Any) -> Any:
    """Encode a value for HDF5 file format.

    Parameters
    ----------
    value
        Value to encode.

    Returns
    -------
    Encoded value.
    """
    if value is None:
        output = "__none__"
    else:
        output = value
    return output


def add_dict_to_hdf5_file(hdf5_file: h5py.File, path: str, dictionary):
    """Save a dictionary to a HDF5 file.

    Based on :code:`recursively_save_dict_contents_to_group` in bilby.

    Parameters
    ----------
    hdf5_file
        HDF5 file.
    path
        Path added to the keys of the dictionary.
    dictionary
        The dictionary to save.
    """
    for key, value in dictionary.items():
        if isinstance(value, dict):
            add_dict_to_hdf5_file(hdf5_file, path + key + "/", value)
        else:
            hdf5_file[path + key] = encode_for_hdf5(value)


def save_dict_to_hdf5(dictionary: dict, filename: str) -> None:
    """Save a dictionary to a HDF5 file.

    Parameters
    ----------
    dictionary
        Dictionary to save.
    filename
        Filename (with the extension) to save the dictionary to. Should include
        the complete path.
    """
    with h5py.File(filename, "w") as f:
        add_dict_to_hdf5_file(f, "/", dictionary)



import logging
import os
import time
from typing import Callable, Optional

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
import torch

from .evidence import EvidenceIntegral
from .plot import (
    plot_trace,
    plot_insertion_indices,
    save_figure,
    plot_samples_1d,
)
from .proposal.base import ProposalWithPool
from .proposal.flow import FlowProposal
from .proposal.prior import PriorProposal
from .utils.sample import rejection_sample
from .utils.stats import rolling_mean_numpy
from .utils.io import save_dict_to_hdf5
from .tensorlist import TensorList


logger = logging.getLogger(__name__)


class Sampler:
    def __init__(
        self,
        log_likelihood: Callable,
        prior_transform: Callable,
        dims: int,
        nlive: int = 1000,
        tolerance: float = 0.1,
        sample_prior_iterations: int = 2000,
        outdir: Optional[str] = None,
        save: bool = True,
        parameter_labels: Optional[list[str]] = None,
        plot_pool: bool = False,
        plot_trace: bool = True,
        plot_insertion_indices: bool = True,
        plot_state: bool = True,
        proposal_class: Optional[Callable] = None,
        reset_flow: int = 0,
        device: str = "cpu",
        seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        self.log_likelihood = log_likelihood
        self.prior_transform = prior_transform
        self.nlive = nlive
        self.dims = dims
        self.tolerance = tolerance
        self.iteration = 0
        self.outdir = os.getcwd() if outdir is None else outdir
        self.save = save
        self.plot_pool = plot_pool
        self._plot_trace = plot_trace
        self._plot_insertion_indices = plot_insertion_indices
        self._plot_state = plot_state
        self.parameter_labels = parameter_labels
        self.seed = seed

        self.reset_flow = int(reset_flow)
        self.sample_prior_iterations = sample_prior_iterations
        self.populate_count = 0
        self.n_likelihood_calls = 0
        self.sampling_time = 0
        self.device = torch.device(device)

        logger.info(f"Running with device={self.device}")

        os.makedirs(self.outdir, exist_ok=True)

        if self.seed:
            logger.info(f"Setting random seed to: {self.seed}")
            torch.manual_seed(self.seed)

        self._nested_samples = TensorList(
            size=(self.dims,), device=self.device
        )
        self.history = dict(
            acceptance=[],
            proposal_acceptance=[],
        )
        self.indices = TensorList(device=self.device, dtype=torch.int)
        self._logl_nested_samples = TensorList(device=self.device)
        self.live_points = None
        self.logl = None
        self.logl_min = -torch.inf
        self.logl_max = -torch.inf
        self.criterion = torch.inf
        self.finalised = False

        self.integral = EvidenceIntegral(nlive=self.nlive, device=self.device)

        if proposal_class is None:
            proposal_class = FlowProposal

        if proposal_class.has_pool and "poolsize" not in kwargs:
            kwargs["poolsize"] = self.nlive

        self.proposal = proposal_class(
            dims=self.dims,
            device=self.device,
            log_likelihood_fn=self.log_likelihood_unit_hypercube,
            **kwargs,
        )

        self.prior_proposal = PriorProposal(
            dims=self.dims,
            device=self.device,
            log_likelihood_fn=self.log_likelihood_unit_hypercube,
        )

    @property
    def stop(self) -> bool:
        if self.criterion < self.tolerance:
            return True
        else:
            return False

    @property
    def nested_samples(self) -> torch.Tensor:
        return self.prior_transform(self._nested_samples.data)

    @property
    def log_likelihoods_nested_samples(self) -> torch.Tensor:
        return self._logl_nested_samples.data

    @property
    def posterior_samples(self) -> torch.Tensor:
        return rejection_sample(
            self.nested_samples,
            self.log_posterior_weights,
        )

    @property
    def log_posterior_weights(self) -> torch.Tensor:
        return self.integral.log_posterior_weights

    @property
    def log_evidence(self) -> torch.Tensor:
        return self.integral.logz

    @property
    def log_evidence_error(self) -> torch.Tensor:
        return self.integral.logz_error

    @property
    def insertion_indices(self) -> torch.Tensor:
        return self.indices.data

    @property
    def should_reset(self) -> bool:
        """Boolean that indicates if the proposal should be reset"""
        return self.reset_flow and not bool(
            self.populate_count % self.reset_flow
        )

    def log_likelihood_unit_hypercube(self, x: torch.Tensor) -> torch.Tensor:
        """Log-likelihood for samples in the unit hypercube"""
        self.n_likelihood_calls += len(x)
        return self.log_likelihood(self.prior_transform(x))

    def initialise(self) -> None:
        live_points, logl = list(
            zip(*[self.prior_proposal.draw(None) for _ in range(self.nlive)])
        )
        live_points = torch.cat(live_points, dim=0)
        logl = torch.cat(logl)
        idx = torch.argsort(logl)
        self.live_points = live_points[idx]
        self.logl = logl[idx]

    def finalise(self) -> None:
        """Finalise the nested sampling run.

        This includes consuming the final live points
        """
        for (i, x), logl in zip(enumerate(self.live_points), self.logl):
            self.integral.update(logl, self.nlive - i)
            self._nested_samples.append(x.detach())
        self.live_points = None

        self.plot()

        logger.info(
            f"Final log-evidence: {self.integral.logz:.3f} +/- "
            f"{self.integral.logz_error:.3f}"
        )

        self.finalised = True

    def insert_live_point(self, x, logl):
        """
        Insert a live point
        """
        # This is the index including the current worst point, so final index
        # is one less, otherwise index=0 would never be possible
        index = torch.searchsorted(self.logl, logl)
        self.live_points[: index - 1, :] = self.live_points[1:index, :].clone()
        self.live_points[index - 1, :] = x
        self.logl.data[: index - 1] = self.logl[1:index].clone()
        self.logl[index - 1] = logl
        return index - 1

    def update_proposal(self) -> None:
        """Update the proposal.

        This includes training and populating the proposal.
        """
        if self.proposal.has_pool:
            if not self.proposal.populated:
                if self.proposal.trainable:
                    with torch.enable_grad():
                        self.proposal.train(
                            self.live_points,
                            self.logl,
                            reset=self.should_reset,
                        )
                proposal_acceptance = self.proposal.populate(
                    self.live_points,
                    self.logl,
                )
                self.history["proposal_acceptance"].append(
                    (self.iteration, proposal_acceptance)
                )
                self.proposal.compute_likelihoods()
                if self.plot_pool:
                    plot_samples_1d(
                        self.live_points,
                        self.proposal.samples,
                        labels=["live points", "pool"],
                        parameter_labels=self.parameter_labels,
                        filename=os.path.join(
                            self.outdir, f"pool_it_{self.iteration}.png"
                        ),
                    )
                self.populate_count += 1
        elif self.proposal.trainable:
            with torch.enable_grad():
                self.proposal.train(
                    self.live_points,
                    self.logl,
                )

    def step(self) -> None:
        """Perform one nested sampling step"""
        self.logl_min = self.logl[0].clone()
        self.integral.update(self.logl_min, self.nlive)
        self._nested_samples.append(self.live_points[0].detach().clone())
        self._logl_nested_samples.append(self.logl[0].detach().clone())
        count = 0
        while True:
            if self.iteration < self.sample_prior_iterations:
                x, logl = self.prior_proposal.draw(self.live_points[0])
            else:
                self.update_proposal()
                x, logl = self.proposal.draw(self.live_points[0])
            if logl is None:
                logl = self.log_likelihood_unit_hypercube(x)
            count += 1
            if torch.isfinite(logl) and logl > self.logl_min:
                break
        index = self.insert_live_point(x, logl)
        self.indices.append(index)
        self.history["acceptance"].append(1 / count)
        self.logl_max = self.logl[-1]
        self.logl_min = self.logl[0]

    def update_criterion(self) -> None:
        self.criterion = (
            torch.logaddexp(
                self.integral.logz, self.logl_max - self.iteration / self.nlive
            )
            - self.integral.logz
        )

    def plot(self) -> None:
        if self._plot_trace:
            plot_trace(
                self.integral.logx.data[1:],
                self.nested_samples,
                labels=self.parameter_labels,
                filename=os.path.join(self.outdir, "trace.png"),
            )

        if self._plot_insertion_indices:
            plot_insertion_indices(
                self.indices.data,
                nlive=self.nlive,
                filename=os.path.join(self.outdir, "insertion_indices.png"),
            )

        if self._plot_state:
            self.plot_state(os.path.join(self.outdir, "state.png"))

    def plot_state(self, filename: Optional[str] = None) -> Optional[Figure]:
        n_plots = 3

        fig, axs = plt.subplots(n_plots, 1, sharex=True)

        its = np.arange(self.iteration)

        j = 0

        # logl will be longer than its after finalizing the run
        axs[j].plot(self.integral.logl.data[1:].cpu().numpy())
        axs[j].set_ylabel(r"$\log L^{*}$")
        j += 1

        axs[j].plot(its, self.history["acceptance"], marker=",", ls="")
        axs[j].plot(
            its,
            rolling_mean_numpy(self.history["acceptance"], self.nlive // 10),
            color="C0",
        )
        axs[j].set_ylabel("Acceptance")
        axs[j].set_yscale("log")
        j += 1
        if len(self.history["proposal_acceptance"]):
            proposal_its, proposal_acceptance = np.array(
                self.history["proposal_acceptance"]
            ).T
            axs[j].plot(proposal_its, proposal_acceptance, marker=".", ls="")
        axs[j].set_ylabel("Proposal acceptance")
        axs[j].set_yscale("log")

        axs[-1].set_xlabel("Iteration")
        return save_figure(fig, filename)

    @torch.no_grad()
    def run(self) -> None:
        start = time.perf_counter()
        self.initialise()

        while not self.stop:
            self.step()
            self.iteration += 1
            self.update_criterion()
            if self.iteration % (self.nlive // 10) == 0:
                logger.info(
                    f"it {self.iteration} - "
                    f"log Z={self.integral.logz:.2f} +/- "
                    f"{self.integral.logz_error:.2f}, "
                    f"log dZ={self.criterion:.2f}, "
                    f"H={self.integral.info:.2f}"
                )
            if self.iteration % (self.nlive) == 0:
                self.plot()

        self.finalise()
        stop = time.perf_counter()
        self.sampling_time += stop - start
        logger.info(f"Total likelihood evaluations: {self.n_likelihood_calls}")
        logger.info(f"Total sampling time : {self.sampling_time:.1f} s")

        if self.save:
            self.save_result()

    def get_result_dictionary(self) -> dict:
        """Return a dictionary containing the results"""
        results = dict()
        results["nested_samples"] = self.nested_samples
        results["log_likelihoods"] = self.log_likelihoods_nested_samples
        results["log_posterior_weights"] = self.log_posterior_weights
        results["log_evidence"] = self.log_evidence
        results["log_evidence_error"] = self.log_evidence_error
        results["criterion"] = self.criterion
        results["insertion_indices"] = self.insertion_indices
        results["n_likelihood_calls"] = self.n_likelihood_calls
        results["sampling_time"] = self.sampling_time
        return results

    def save_result(self, filename: Optional[str] = None) -> None:
        """Save the results to a file.

        Defaults to :code:`result.hdf5` if the filename is not specified.
        """
        if filename is None:
            filename = os.path.join(self.outdir, "result.hdf5")
        save_dict_to_hdf5(self.get_result_dictionary(), filename)
